# -*- coding: utf-8 -*-
# Description:
# Support for PaloAlto API.
#
# Authors:
# Benoît Allard <benoit.allard@greenbone.net>
#
# Copyright:
# Copyright (C) 2015 Greenbone Networks GmbH
#
# This program is free software; you can redistribute it and/or
# modify it under the terms of the GNU General Public License
# as published by the Free Software Foundation; either version 2
# of the License, or (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA 02110-1301 USA.

import datetime
import logging
import os
from requests import get, exceptions
import tempfile
import xml.etree.ElementTree as ET

from ospd.ospd import OSPDError

logger = logging.getLogger(__name__)

class APIError(Exception):
    def __init__(self, stuff):
        self.stuff = stuff
    def __str__(self):
        return ET.tostring(self.stuff)


def make_members(parent, values):
    for val in values:
        member = ET.SubElement(parent, 'member')
        member.text = val


class PANOS410(object):
    
    """
    This is a client implementation of the PAN-OS 4.1.0 REST API
    """

    def __init__(self, hostname):
        self.hostname = hostname
        self.ca = None
        self.key = None

    def __del__(self):
        """ delete the ca file if it is there """
        if self.ca is None:
            return
        if not os.path.isfile(self.ca):
            return
        try:
            os.unlink(self.ca)
        except OSError:
            pass

    def setCA(self, certificate):
        fd, name = tempfile.mkstemp()
        self.ca = name
        os.write(fd, certificate)
        os.close(fd)

    def _get(self, type_, params=None, addKey=True):
        if params is None:
            params = {}
        params.update({'type': type_})
        if addKey and self.key:
            params.update({'key': self.key})
        verify = False
        if self.ca is not None:
            verify=self.ca
        r = get('https://%s/api/' % self.hostname,
                params=params,
                verify=verify)
        #logger.debug(r.url)
        root = ET.fromstring(r.text)
        if (root.tag != 'response' or root.attrib.get('status') != 'success') and (root.tag != 'report'):
            raise APIError(root)
        #logger.debug(ET.tostring(root))
        return root
        
    def keygen(self, username, password):
        """
            2.1 Key generation
        """
        xml = self._get('keygen', dict(user=username, password=password), addKey=False)
        self.key = xml.findtext('result/key')
        return self.key is not None

    def _report(self, type_, name, period=None, cmd=None):
        """
            2.5 Reporting
        """
        params = dict(reporttype=type_, reportname=name)
        if (type_ == 'dynamic') and (name == 'custom-dynamic-report') and (cmd is not None):
            # Join the XML elements together without the enclosing element.
            cmd = ''.join([ET.tostring(scmd) for scmd in cmd])
            params.update(dict(cmd=cmd))
        elif (type_ == 'dynamic') and (period is not None):
            params.update(dict(period=period))
        xml = self._get('report', params)
        return xml.find('report')
    
    def _config(self, action, xpath):
        return self._get('config', dict(action=action, xpath=xpath))

    def get_config(self, xpath):
        return self._config('get', xpath)

    def get_report_from_config(self, xpath, index=0):
        reportsdef = self.get_config(xpath)
        reportdef = reportsdef.findall('result/entry')[index]
        self._report('dynamic', 'custom-dynamic-report', cmd=reportdef)

    def custom_report(self, type_, agg_values, values, period, query, nresults, group_by=None, ngroups=0):
        reportdef = ET.Element('reportdef')
        reporttype = ET.SubElement(reportdef, 'type')
        report = ET.SubElement(reporttype, type_)
        if group_by:
            grp_by = ET.SubElement(report, 'group-by')
            grp_by.text = group_by
        aggregate_by = ET.SubElement(report, 'aggregate-by')
        make_members(aggregate_by, agg_values)
        if values:
            valueselem = ET.SubElement(report, 'values')
            make_members(valueselem, values)
        period_ = ET.SubElement(reportdef, 'period')
        period_.text = period
        query_ = ET.SubElement(reportdef, 'query')
        query_.text = query
        # Amount of results
        topn = ET.SubElement(reportdef, 'topn')
        topn.text = str(nresults)
        if group_by:
            # Amount of groups
            topm = ET.SubElement(reportdef, 'topm')
            topm.text = str(ngroups)
        logger.debug(ET.tostring(reportdef))

        result = self._report('dynamic', 'custom-dynamic-report', cmd=reportdef)
        for res in result.findall('result/entry'):
            yield res


class PANOSClient(PANOS410):
    """
        High-Level Client for the PAN-OS
    """

    def __init__(self, hostname, username, password, ca=None):
        PANOS410.__init__(self, hostname)
        if ca:
            self.setCA(ca)
        if not self.keygen(username, password):
            raise OSPDError("Unable to connect to the paloalto appliance.")
        # A cache for the already-fetched vulnerabilities
        self.vulns = {}

    def get_vulnerability(self, name):
        if name not in self.vulns:
            xml = self.get_config("/config/predefined/threats/vulnerability/entry[@name='%s']" % name)
            entry = xml.find('result/entry')
            if entry is None:
                return None
            self.vulns[name] = Vulnerability(entry)
        return self.vulns[name]

    def get_threats(self, address, period, group_by=None, nresults=500):
        for threat in self.custom_report(Threat.reportname,
                Threat.aggregates,
                Threat.values,
                period,
                '(addr.dst in %s)' % address,
                nresults,
                group_by,
                50):
            yield Threat(threat)

    def get_traffic(self, address, period, nresults=1000):
        # We should add a check that we got less than the required amount of
	# traffic in order to be sure that we got it all.
        for traffic in self.custom_report(Traffic.reportname,
                Traffic.aggregates,
                Traffic.values,
                period,
                '(addr.dst in %s) and (bytes_received geq 1)' % address,
                nresults):
            yield Traffic(traffic)


class Vulnerability(object):
    """ A Vulnerability as PaloAlto know them
    """
    def __init__(self, xml):
        self.name = int(xml.attrib['name'])
        self.threatname = xml.findtext('threatname')
        
        self.cves = [e.text for e in xml.findall('cve/member')]
        self.vendors = [e.text for e in xml.findall('vendor/member')]
        self.severity = xml.findtext('severity')

    def __str__(self):
        return '%s (%s): %s [%s]' % (self.threatname, self.name, ', '.join(self.cves), self.severity)
        

class Threat(object):
    """ This object represents a Threat from the 'threat' Table
    """
    # Those three values are for the query
    reportname = 'threat'
    aggregates = ('threatid', 'severity', 'src', 'sport', 'dst', 'dport', 'proto', 'action')
    values = ('repeatcnt',)
    def __init__(self, xml):
        self.name = xml.findtext('threatid')
        self.id = int(xml.findtext('tid', 0))
        self.source = (xml.findtext('src'), int(xml.findtext('sport', 0)))
        self.destination = (xml.findtext('dst'), int(xml.findtext('dport', 0)))
        self.severity = xml.findtext('severity')
        self.protocol = xml.findtext('proto')
        self.action = xml.findtext('action')
        self.count = int(xml.findtext('repeatcnt', 0))

    def __str__(self):
        return "%s: %s (%d) [%s -> %s]" % (self.tstamp, self.name, self.id, self.source, self.destination)

    def __eq__(self, other):
        return (self.id, self.destination, self.protocol) == (other.id, other.destination, other.protocol)

    def __hash__(self):
        """ An indexing key """
        return hash((self.id, self.destination[1], self.protocol))
        

class Traffic(object):
    """ One Traffic as logged by the PaloAlto.
    """
    # Same as for the Threat, those are the queried fields.
    reportname = 'traffic'
    aggregates = ('src', 'sport', 'dst', 'dport', 'proto', 'app', 'risk-of-app')
    values = () # ('bytes_received',)
    def __init__(self, xml):
        self.src = (xml.findtext('src'), int(xml.findtext('sport')))
        self.dst = (xml.findtext('dst'), int(xml.findtext('dport')))
        self.protocol = xml.findtext('proto')
        self.app = (xml.findtext('app'), xml.findtext('risk-of-app'))

    def __str__(self):
        return "%s:%s -> %s:%s (%s)" % (self.src[0], self.src[1], self.dst[0], self.dst[1], self.protocol)
